#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2021 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################
from __future__ import annotations

import time
import datetime as DT
from dateutil import tz
from dataclasses import dataclass, fields, field
from typing import List, Optional
import nardascripting
from nardascripting.base.signalsharkdev import *
from nardascripting.base.usrscriptbase import *
from nardascripting.base.scpi.enums import *
# import nardascripting.tools.xmlhandler as xml


def wait_for_new_data(signalshark: SignalSharkDev, trials: int = 3, t_wait: float = 1.0) -> bool:
    """Queries data_update and blocks, until new data is available.

    :param signalshark: The SignalShark device.
    :type signalshark: nardascripting.base.signalsharkdev.SignalSharkDev
    :param trials: Number of trials.
    :type trials: int
    :param t_wait: Time to wait between trials.
    :type t_wait: float
    :return: Returns true if succeeded.
    :rtype: bool
    """
    try:
        trial_cnt = 0
        while True:
            # Wait until Scan Number has increased.
            scan_number = signalshark.scpi.spectrum.get_data_update()
            if scan_number > 0:
                return True
            else:
                trial_cnt += 1
                if trial_cnt < trials:
                    time.sleep(t_wait)
                    continue
                else:
                    return False
    except Exception:
        return False


def do_auto_atten(signalshark: SignalSharkDev, start_atten: int = 0, samples: int = 3) -> float:
    """Sets the attenuation automatically.

    :param signalshark: The SignalShark device.
    :type signalshark: nardascripting.base.signalsharkdev.SignalSharkDev
    :param start_atten: ttenuation value to start with.
    :type start_atten: float
    :param samples: Number of samples to check overdriven flag.
    :type samples: int
    :return: Returns the set attenuation or -1.0.
    :rtype: float
    """
    try:
        if not signalshark.connect():
            return -1.0

        for atten in range(start_atten, 34, 3):
            if atten > 31.5:
                atten = 31.5
            signalshark.scpi.sense.set_attenuator(atten)
            overdriven = False
            for sample in range(samples):
                if not wait_for_new_data(signalshark):
                    return -1.0
                overdriven |= signalshark.scpi.spectrum.get_data_overdriven()
            if not overdriven:
                break
        return signalshark.scpi.sense.get_attenuator()
    except Exception:
        return -1.0


@dataclass
class SpecSettings(object):
    """Class that contains information about spectrum settings."""
    name: str = field(default="Messung 1", metadata={"XmlName": "Name"})
    comment: str = field(default="", metadata={"XmlName": "Comment"})

    fstart: float = field(default=80.0, metadata={"XmlName": "Fstart", "Unit": "Hz"})
    fstop: float = field(default=110.0, metadata={"XmlName": "Fstop", "Unit": "Hz"})
    fstep: float = field(default=0.0, metadata={"XmlName": "Fstep", "Unit": "Hz"})
    bin_cnt: int = field(default=0, metadata={"XmlName": "BinCnt"})

    trace_list: List[LevDetectorModes] = field(default_factory=lambda: [], metadata={"XmlName": "TraceList"})
    trace_infinite: bool = field(default=False, metadata={"XmlName": "TraceInfinite"})
    unit: LevelUnits = field(default=LevelUnits.dBm, metadata={"XmlName": "UnitLevel"})
    ymax: float = field(default=80.0, metadata={"XmlName": "YMax"})
    yrange: float = field(default=80.0, metadata={"XmlName": "YRange"})
    rbw: float = field(default=0.01, metadata={"XmlName": "RBW", "Unit": "Hz"})
    rbw_filter_type: RbwFilterTypes = field(default=RbwFilterTypes.NORMAL,
                                            metadata={"XmlName": "RbwFilterType"})
    meas_time: float = field(default=0.02, metadata={"XmlName": "MeasTime", "Unit": "s"})
    auto_atten: bool = field(default=False, metadata={"XmlName": "AutoAtten"})
    atten: float = field(default=0.0, metadata={"XmlName": "Atten", "Unit": "dB"})

    def get_from_device(self, signalshark: SignalSharkDev) -> (bool, str):
        """Gets all spectrum dedicated parameters form the given SignalShark.

        :param signalshark: The SignalShark device.
        :type signalshark: nardascripting.base.signalsharkdev.SignalSharkDev
        :return: Returns True/False for passed/failed and a message text as Tuple.
        :rtype: (bool, str)
        """
        err_msg = "Can not read the parameters from the device!"
        try:
            # Connect to SignalShark
            if not signalshark.connect():
                return False, "Connection to the device cannot be established!"

            # Clear SCPI errors:
            signalshark.scpi.check_error()

            # Readout settings:
            self.fstart = signalshark.scpi.spectrum.get_frequency_start()
            self.fstop = signalshark.scpi.spectrum.get_frequency_stop()
            self.fstep = signalshark.scpi.spectrum.get_data_frequency_step()
            self.bin_cnt = signalshark.scpi.spectrum.get_data_count()
            self.trace_list = signalshark.scpi.spectrum.get_trace_list()
            self.trace_infinite = signalshark.scpi.spectrum.get_trace_infinite()
            self.unit = signalshark.scpi.display.get_unit()
            self.ymax = signalshark.scpi.display.get_spectrum_lmax()
            self.yrange = signalshark.scpi.display.get_spectrum_lrange()
            self.rbw = signalshark.scpi.spectrum.get_rbw()
            self.rbw_filter_type = signalshark.scpi.spectrum.get_rbw_filter_type()
            self.meas_time = signalshark.scpi.spectrum.get_measurement_time()
            self.atten = signalshark.scpi.sense.get_attenuator()

            # Check for SCPI errors:
            if signalshark.scpi.check_no_error():
                return True, ""
            else:
                return False, err_msg
        except Exception as e:
            return False, err_msg

    def set_to_device(self, signalshark: SignalSharkDev) -> (bool, str):
        """Sets all spectrum dedicated parameters to the given SignalShark.

        :param signalshark: The SignalShark device.
        :type signalshark: nardascripting.base.signalsharkdev.SignalSharkDev
        :return: Returns True/False for passed/failed and a message text as Tuple.
        :rtype: (bool, str)
        """
        err_msg = "Can not write the parameters to the device!"
        try:
            # Connect to SignalShark
            if not signalshark.connect():
                return False, "Connection to the device cannot be established!"

            # Clear SCPI errors:
            signalshark.scpi.check_error()

            # Write settings:
            signalshark.scpi.spectrum.set_frequency_stop(self.fstop)
            signalshark.scpi.spectrum.get_frequency_start(self.fstart)

            for item in LevDetectorModes:
                signalshark.scpi.spectrum.set_trace_enable(item, item in self.trace_list)
            signalshark.scpi.spectrum.set_trace_infinite(self.trace_infinite)

            signalshark.scpi.display.set_unit(self.unit)
            signalshark.scpi.display.set_spectrum_lmax(self.ymax)
            signalshark.scpi.display.set_spectrum_lrange(self.yrange)

            signalshark.scpi.spectrum.set_rbw(self.rbw)
            signalshark.scpi.spectrum.get_rbw_filter_type(self.rbw_filter_type)
            signalshark.scpi.spectrum.get_measurement_time(self.meas_time)

            if self.auto_atten:
                do_auto_atten(signalshark)
            else:
                signalshark.scpi.sense.set_attenuator(self.atten)
                wait_for_new_data(signalshark)

            # Check for SCPI errors:
            if signalshark.scpi.check_no_error():
                return True, ""
            else:
                return False, err_msg
        except Exception as e:
            return False, err_msg

    # def save_to_file(self, filename: str) -> bool:
    #     """Saves spectrum settings to XML file.
    #
    #     :param filename: The name of the settings file.
    #     :type filename: str
    #     :return: Returns true if succeeded.
    #     """
    #     return xml.save_as_xml_file(self, filename)
    #
    # @staticmethod
    # def load_from_file(filename: str) -> Optional[SpecSettings]:
    #     """Loads spectrum settings from XML file.
    #
    #     :param filename: The name of the settings file.
    #     :type filename: str
    #     :return: Returns the SpecSettings object or None
    #     """
    #     return xml.load_from_xml_file(SpecSettings, filename)


# This is just a python module containing helper classes/functions:
if __name__ == '__main__':
    pass
